@@ -448,7 +448,7 @@ parse_event(void)
     tok = ENDINPUT;
     incmdpos = 1;
     aliasspaceflag = 0;
-    yylex();
+    zshlex();
     init_parse();
 
     if (!par_event()) {
@@ -467,7 +467,7 @@ par_event(void)
     while (tok == SEPER) {
 	if (isnewlin > 0)
 	    return 0;
-	yylex();
+	zshlex();
     }
     if (tok == ENDINPUT)
 	return 0;
@@ -481,15 +481,15 @@ par_event(void)
 	} else if (tok == SEPER) {
 	    set_list_code(p, Z_SYNC, c);
 	    if (isnewlin <= 0)
-		yylex();
+		zshlex();
 	    r = 1;
 	} else if (tok == AMPER) {
 	    set_list_code(p, Z_ASYNC, c);
-	    yylex();
+	    zshlex();
 	    r = 1;
 	} else if (tok == AMPERBANG) {
 	    set_list_code(p, (Z_ASYNC | Z_DISOWN), c);
-	    yylex();
+	    zshlex();
 	    r = 1;
 	}
     }
@@ -525,7 +525,7 @@ parse_list(void)
 
     tok = ENDINPUT;
     incmdpos = 1;
-    yylex();
+    zshlex();
     init_parse();
     par_list(&c);
     if (tok != ENDINPUT) {
@@ -601,7 +601,7 @@ par_list(int *complex)
  rec:
 
     while (tok == SEPER)
-	yylex();
+	zshlex();
 
     p = ecadd(0);
     c = 0;
@@ -616,7 +616,7 @@ par_list(int *complex)
 			      (Z_ASYNC | Z_DISOWN)), c);
 	    incmdpos = 1;
 	    do {
-		yylex();
+		zshlex();
 	    } while (tok == SEPER);
 	    lp = p;
 	    goto rec;
@@ -669,9 +669,9 @@ par_sublist(int *complex)
 	    int qtok = tok, sl;
 
 	    cmdpush(tok == DBAR ? CS_CMDOR : CS_CMDAND);
-	    yylex();
+	    zshlex();
 	    while (tok == SEPER)
-		yylex();
+		zshlex();
 	    sl = par_sublist(complex);
 	    set_sublist_code(p, (sl ? (qtok == DBAR ?
 				       WC_SUBLIST_OR : WC_SUBLIST_AND) :
@@ -700,11 +700,11 @@ par_sublist2(int *complex)
     if (tok == COPROC) {
 	*complex = 1;
 	f |= WC_SUBLIST_COPROC;
-	yylex();
+	zshlex();
     } else if (tok == BANG) {
 	*complex = 1;
 	f |= WC_SUBLIST_NOT;
-	yylex();
+	zshlex();
     }
     if (!par_pline(complex) && !f)
 	return -1;
@@ -732,9 +732,9 @@ par_pline(int *complex)
     if (tok == BAR) {
 	*complex = 1;
 	cmdpush(CS_PIPE);
-	yylex();
+	zshlex();
 	while (tok == SEPER)
-	    yylex();
+	    zshlex();
 	ecbuf[p] = WCB_PIPE(WC_PIPE_MID, (line >= 0 ? line + 1 : 0));
 	ecispace(p + 1, 1);
 	ecbuf[p + 1] = ecused - 1 - p;
@@ -756,9 +756,9 @@ par_pline(int *complex)
 
 	*complex = 1;
 	cmdpush(CS_ERRPIPE);
-	yylex();
+	zshlex();
 	while (tok == SEPER)
-	    yylex();
+	    zshlex();
 	ecbuf[p] = WCB_PIPE(WC_PIPE_MID, (line >= 0 ? line + 1 : 0));
 	ecispace(p + 1, 1);
 	ecbuf[p + 1] = ecused - 1 - p;
@@ -856,7 +856,7 @@ par_cmd(int *complex)
     case DINPAR:
 	ecadd(WCB_ARITH());
 	ecstr(tokstr);
-	yylex();
+	zshlex();
 	break;
     case TIME:
 	{
@@ -917,23 +917,23 @@ par_for(int *complex)
 
     incmdpos = 0;
     infor = tok == FOR ? 2 : 0;
-    yylex();
+    zshlex();
     if (tok == DINPAR) {
-	yylex();
+	zshlex();
 	if (tok != DINPAR)
 	    YYERRORV(oecused);
 	ecstr(tokstr);
-	yylex();
+	zshlex();
 	if (tok != DINPAR)
 	    YYERRORV(oecused);
 	ecstr(tokstr);
-	yylex();
+	zshlex();
 	if (tok != DOUTPAR)
 	    YYERRORV(oecused);
 	ecstr(tokstr);
 	infor = 0;
 	incmdpos = 1;
-	yylex();
+	zshlex();
 	type = WC_FOR_COND;
     } else {
 	int np = 0, n, posix_in, ona = noaliases, onc = nocorrect;
@@ -948,7 +948,7 @@ par_for(int *complex)
 	for (;;) {
 	    n++;
 	    ecstr(tokstr);
-	    yylex();
+	    zshlex();
 	    if (tok != STRING || !strcmp(tokstr, "in") || sel)
 		break;
 	    if (!isident(tokstr) || errflag)
@@ -964,10 +964,10 @@ par_for(int *complex)
 	    ecbuf[np] = n;
 	posix_in = isnewlin;
 	while (isnewlin)
-	    yylex();
+	    zshlex();
         if (tok == STRING && !strcmp(tokstr, "in")) {
 	    incmdpos = 0;
-	    yylex();
+	    zshlex();
 	    np = ecadd(0);
 	    n = par_wordlist();
 	    if (tok != SEPER)
@@ -976,38 +976,38 @@ par_for(int *complex)
 	    type = (sel ? WC_SELECT_LIST : WC_FOR_LIST);
 	} else if (!posix_in && tok == INPAR) {
 	    incmdpos = 0;
-	    yylex();
+	    zshlex();
 	    np = ecadd(0);
 	    n = par_nl_wordlist();
 	    if (tok != OUTPAR)
 		YYERRORV(oecused);
 	    ecbuf[np] = n;
 	    incmdpos = 1;
-	    yylex();
+	    zshlex();
 	    type = (sel ? WC_SELECT_LIST : WC_FOR_LIST);
 	} else
 	    type = (sel ? WC_SELECT_PPARAM : WC_FOR_PPARAM);
     }
     incmdpos = 1;
     while (tok == SEPER)
-	yylex();
+	zshlex();
     if (tok == DOLOOP) {
-	yylex();
+	zshlex();
 	par_save_list(complex);
 	if (tok != DONE)
 	    YYERRORV(oecused);
-	yylex();
+	zshlex();
     } else if (tok == INBRACE) {
-	yylex();
+	zshlex();
 	par_save_list(complex);
 	if (tok != OUTBRACE)
 	    YYERRORV(oecused);
-	yylex();
+	zshlex();
     } else if (csh || isset(CSHJUNKIELOOPS)) {
 	par_save_list(complex);
 	if (tok != ZEND)
 	    YYERRORV(oecused);
-	yylex();
+	zshlex();
     } else if (unset(SHORTLOOPS)) {
 	YYERRORV(oecused);
     } else
@@ -1035,7 +1035,7 @@ par_case(int *complex)
     p = ecadd(0);
 
     incmdpos = 0;
-    yylex();
+    zshlex();
     if (tok != STRING)
 	YYERRORV(oecused);
     ecstr(tokstr);
@@ -1044,9 +1044,9 @@ par_case(int *complex)
     ona = noaliases;
     onc = nocorrect;
     noaliases = nocorrect = 1;
-    yylex();
+    zshlex();
     while (tok == SEPER)
-	yylex();
+	zshlex();
     if (!(tok == STRING && !strcmp(tokstr, "in")) && tok != INBRACE)
     {
 	noaliases = ona;
@@ -1058,17 +1058,17 @@ par_case(int *complex)
     incmdpos = 0;
     noaliases = ona;
     nocorrect = onc;
-    yylex();
+    zshlex();
 
     for (;;) {
 	char *str;
 
 	while (tok == SEPER)
-	    yylex();
+	    zshlex();
 	if (tok == OUTBRACE)
 	    break;
 	if (tok == INPAR)
-	    yylex();
+	    zshlex();
 	if (tok != STRING)
 	    YYERRORV(oecused);
 	if (!strcmp(tokstr, "esac"))
@@ -1078,11 +1078,11 @@ par_case(int *complex)
 	incmdpos = 1;
 	type = WC_CASE_OR;
 	for (;;) {
-	    yylex();
+	    zshlex();
 	    if (tok == OUTPAR) {
 		incasepat = 0;
 		incmdpos = 1;
-		yylex();
+		zshlex();
 		break;
 	    } else if (tok == BAR) {
 		char *str2;
@@ -1157,10 +1157,10 @@ par_case(int *complex)
 	    YYERRORV(oecused);
 	incasepat = 1;
 	incmdpos = 0;
-	yylex();
+	zshlex();
     }
     incmdpos = 1;
-    yylex();
+    zshlex();
 
     ecbuf[p] = WCB_CASE(WC_CASE_HEAD, ecused - 1 - p);
 }
@@ -1184,13 +1184,13 @@ par_if(int *complex)
     for (;;) {
 	xtok = tok;
 	cmdpush(xtok == IF ? CS_IF : CS_ELIF);
-	yylex();
+	zshlex();
 	if (xtok == FI)
 	    break;
 	if (xtok == ELSE)
 	    break;
 	while (tok == SEPER)
-	    yylex();
+	    zshlex();
 	if (!(xtok == IF || xtok == ELIF)) {
 	    cmdpop();
 	    YYERRORV(oecused);
@@ -1199,15 +1199,19 @@ par_if(int *complex)
 	type = (xtok == IF ? WC_IF_IF : WC_IF_ELIF);
 	par_save_list(complex);
 	incmdpos = 1;
+	if (tok == ENDINPUT) {
+	    cmdpop();
+	    YYERRORV(oecused);
+	}
 	while (tok == SEPER)
-	    yylex();
+	    zshlex();
 	xtok = FI;
 	nc = cmdstack[cmdsp - 1] == CS_IF ? CS_IFTHEN : CS_ELIFTHEN;
 	if (tok == THEN) {
 	    usebrace = 0;
 	    cmdpop();
 	    cmdpush(nc);
-	    yylex();
+	    zshlex();
 	    par_save_list(complex);
 	    ecbuf[pp] = WCB_IF(type, ecused - 1 - pp);
 	    incmdpos = 1;
@@ -1216,14 +1220,14 @@ par_if(int *complex)
 	    usebrace = 1;
 	    cmdpop();
 	    cmdpush(nc);
-	    yylex();
+	    zshlex();
 	    par_save_list(complex);
 	    if (tok != OUTBRACE) {
 		cmdpop();
 		YYERRORV(oecused);
 	    }
 	    ecbuf[pp] = WCB_IF(type, ecused - 1 - pp);
-	    yylex();
+	    zshlex();
 	    incmdpos = 1;
 	    if (tok == SEPER)
 		break;
@@ -1245,9 +1249,9 @@ par_if(int *complex)
 	pp = ecadd(0);
 	cmdpush(CS_ELSE);
 	while (tok == SEPER)
-	    yylex();
+	    zshlex();
 	if (tok == INBRACE && usebrace) {
-	    yylex();
+	    zshlex();
 	    par_save_list(complex);
 	    if (tok != OUTBRACE) {
 		cmdpop();
@@ -1261,7 +1265,7 @@ par_if(int *complex)
 	    }
 	}
 	ecbuf[pp] = WCB_IF(WC_IF_ELSE, ecused - 1 - pp);
-	yylex();
+	zshlex();
 	cmdpop();
     }
     ecbuf[p] = WCB_IF(WC_IF_HEAD, ecused - 1 - p);
@@ -1280,28 +1284,28 @@ par_while(int *complex)
     int type = (tok == UNTIL ? WC_WHILE_UNTIL : WC_WHILE_WHILE);
 
     p = ecadd(0);
-    yylex();
+    zshlex();
     par_save_list(complex);
     incmdpos = 1;
     while (tok == SEPER)
-	yylex();
+	zshlex();
     if (tok == DOLOOP) {
-	yylex();
+	zshlex();
 	par_save_list(complex);
 	if (tok != DONE)
 	    YYERRORV(oecused);
-	yylex();
+	zshlex();
     } else if (tok == INBRACE) {
-	yylex();
+	zshlex();
 	par_save_list(complex);
 	if (tok != OUTBRACE)
 	    YYERRORV(oecused);
-	yylex();
+	zshlex();
     } else if (isset(CSHJUNKIELOOPS)) {
 	par_save_list(complex);
 	if (tok != ZEND)
 	    YYERRORV(oecused);
-	yylex();
+	zshlex();
     } else
 	YYERRORV(oecused);
 
@@ -1321,31 +1325,31 @@ par_repeat(int *complex)
     p = ecadd(0);
 
     incmdpos = 0;
-    yylex();
+    zshlex();
     if (tok != STRING)
 	YYERRORV(oecused);
     ecstr(tokstr);
     incmdpos = 1;
-    yylex();
+    zshlex();
     while (tok == SEPER)
-	yylex();
+	zshlex();
     if (tok == DOLOOP) {
-	yylex();
+	zshlex();
 	par_save_list(complex);
 	if (tok != DONE)
 	    YYERRORV(oecused);
-	yylex();
+	zshlex();
     } else if (tok == INBRACE) {
-	yylex();
+	zshlex();
 	par_save_list(complex);
 	if (tok != OUTBRACE)
 	    YYERRORV(oecused);
-	yylex();
+	zshlex();
     } else if (isset(CSHJUNKIELOOPS)) {
 	par_save_list(complex);
 	if (tok != ZEND)
 	    YYERRORV(oecused);
-	yylex();
+	zshlex();
     } else if (unset(SHORTLOOPS)) {
 	YYERRORV(oecused);
     } else
@@ -1368,20 +1372,20 @@ par_subsh(int *complex)
     p = ecadd(0);
     /* Extra word only needed for always block */
     pp = ecadd(0);
-    yylex();
+    zshlex();
     par_list(complex);
     ecadd(WCB_END());
     if (tok != ((otok == INPAR) ? OUTPAR : OUTBRACE))
 	YYERRORV(oecused);
     incmdpos = 1;
-    yylex();
+    zshlex();
 
     /* Optional always block.  No intervening SEPERs allowed. */
     if (otok == INBRACE && tok == STRING && !strcmp(tokstr, "always")) {
 	ecbuf[pp] = WCB_TRY(ecused - 1 - pp);
 	incmdpos = 1;
 	do {
-	    yylex();
+	    zshlex();
 	} while (tok == SEPER);
 
 	if (tok != INBRACE)
@@ -1389,16 +1393,16 @@ par_subsh(int *complex)
 	cmdpop();
 	cmdpush(CS_ALWAYS);
 
-	yylex();
+	zshlex();
 	par_save_list(complex);
 	while (tok == SEPER)
-	    yylex();
+	    zshlex();
 
 	incmdpos = 1;
 
 	if (tok != OUTBRACE)
 	    YYERRORV(oecused);
-	yylex();
+	zshlex();
 	ecbuf[p] = WCB_TRY(ecused - 1 - p);
     } else {
 	ecbuf[p] = (otok == INPAR ? WCB_SUBSH(ecused - 1 - p) :
@@ -1422,7 +1426,7 @@ par_funcdef(void)
     lineno = 0;
     nocorrect = 1;
     incmdpos = 0;
-    yylex();
+    zshlex();
 
     p = ecadd(0);
     ecadd(0);
@@ -1435,7 +1439,7 @@ par_funcdef(void)
 	}
 	ecstr(tokstr);
 	num++;
-	yylex();
+	zshlex();
     }
     ecadd(0);
     ecadd(0);
@@ -1443,9 +1447,9 @@ par_funcdef(void)
 
     nocorrect = 0;
     if (tok == INOUTPAR)
-	yylex();
+	zshlex();
     while (tok == SEPER)
-	yylex();
+	zshlex();
 
     ecnfunc++;
     ecssub = so = ecsoffs;
@@ -1453,7 +1457,7 @@ par_funcdef(void)
     ecnpats = 0;
 
     if (tok == INBRACE) {
-	yylex();
+	zshlex();
 	par_list(&c);
 	if (tok != OUTBRACE) {
 	    lineno += oldlineno;
@@ -1461,7 +1465,7 @@ par_funcdef(void)
 	    ecssub = oecssub;
 	    YYERRORV(oecused);
 	}
-	yylex();
+	zshlex();
     } else if (unset(SHORTLOOPS)) {
 	lineno += oldlineno;
 	ecnpats = onp;
@@ -1494,7 +1498,7 @@ par_time(void)
 {
     int p, f, c = 0;
 
-    yylex();
+    zshlex();
 
     p = ecadd(0);
     ecadd(0);
@@ -1519,13 +1523,13 @@ par_dinbrack(void)
 
     incond = 1;
     incmdpos = 0;
-    yylex();
+    zshlex();
     par_cond();
     if (tok != DOUTBRACK)
 	YYERRORV(oecused);
     incond = 0;
     incmdpos = 1;
-    yylex();
+    zshlex();
 }
 
 /*
@@ -1542,7 +1546,7 @@ static int
 par_simple(int *complex, int nr)
 {
     int oecused = ecused, isnull = 1, r, argc = 0, p, isfunc = 0, sr = 0;
-    int c = *complex, nrediradd;
+    int c = *complex, nrediradd, assignments = 0;
 
     r = ecused;
     for (;;) {
@@ -1574,7 +1578,7 @@ par_simple(int *complex, int nr)
 		 * we need process handling.
 		 */
 		if (p[1] == Inpar &&
-		    (*p == Equals || *p == Inang || *p == Outang)) {
+		    (*p == Equals || *p == Inang || *p == OutangProc)) {
 		    *complex = 1;
 		    break;
 		}
@@ -1582,6 +1586,7 @@ par_simple(int *complex, int nr)
 	    ecstr(name);
 	    ecstr(str);
 	    isnull = 0;
+	    assignments = 1;
 	} else if (tok == ENVARRAY) {
 	    int oldcmdpos = incmdpos, n, type2;
 
@@ -1594,7 +1599,7 @@ par_simple(int *complex, int nr)
 		type2 = WC_ASSIGN_NEW;
 	    ecstr(tokstr);
 	    cmdpush(CS_ARRAY);
-	    yylex();
+	    zshlex();
 	    n = par_nl_wordlist();
 	    ecbuf[p] = WCB_ASSIGN(WC_ASSIGN_ARRAY, type2, n);
 	    cmdpop();
@@ -1602,9 +1607,10 @@ par_simple(int *complex, int nr)
 		YYERROR(oecused);
 	    incmdpos = oldcmdpos;
 	    isnull = 0;
+	    assignments = 1;
 	} else
 	    break;
-	yylex();
+	zshlex();
     }
     if (tok == AMPER || tok == AMPERBANG)
 	YYERROR(oecused);
@@ -1630,7 +1636,7 @@ par_simple(int *complex, int nr)
 			char *toksave = tokstr;
 			char *idstring = dupstrpfx(tokstr+1, eptr-tokstr-1);
 			redir_var = 1;
-			yylex();
+			zshlex();
 
 			if (IS_REDIROP(tok) && tokfd == -1)
 			{
@@ -1652,7 +1658,7 @@ par_simple(int *complex, int nr)
 	    {
 		ecstr(tokstr);
 		argc++;
-		yylex();
+		zshlex();
 	    }
 	} else if (IS_REDIROP(tok)) {
 	    *complex = c = 1;
@@ -1663,16 +1669,20 @@ par_simple(int *complex, int nr)
 	    zlong oldlineno = lineno;
 	    int onp, so, oecssub = ecssub;
 
+	    /* Error if too many function definitions at once */
 	    if (!isset(MULTIFUNCDEF) && argc > 1)
 		YYERROR(oecused);
+	    /* Error if preceding assignments */
+	    if (assignments)
+		YYERROR(oecused);
 
 	    *complex = c;
 	    lineno = 0;
 	    incmdpos = 1;
 	    cmdpush(CS_FUNCDEF);
-	    yylex();
+	    zshlex();
 	    while (tok == SEPER)
-		yylex();
+		zshlex();
 
 	    ecispace(p + 1, 1);
 	    ecbuf[p + 1] = argc;
@@ -1688,7 +1698,7 @@ par_simple(int *complex, int nr)
 	    if (tok == INBRACE) {
 		int c = 0;
 
-		yylex();
+		zshlex();
 		par_list(&c);
 		if (tok != OUTBRACE) {
 		    cmdpop();
@@ -1697,7 +1707,7 @@ par_simple(int *complex, int nr)
 		    ecssub = oecssub;
 		    YYERROR(oecused);
 		}
-		yylex();
+		zshlex();
 	    } else {
 		int ll, sl, pl, c = 0;
 
@@ -1784,7 +1794,7 @@ par_redir(int *rp, char *idstring)
 	nocorrect = 1;
     type = redirtab[tok - OUTANG];
     fd1 = tokfd;
-    yylex();
+    zshlex();
     if (tok != STRING && tok != ENVSTRING)
 	YYERROR(ecused);
     incmdpos = oldcmdpos;
@@ -1803,13 +1813,17 @@ par_redir(int *rp, char *idstring)
 	struct heredocs **hd;
 	int htype = type;
 
+	/*
+	 * Add two here for the string to remember the HERE
+	 * terminator in raw and munged form.
+	 */
 	if (idstring)
 	{
 	    type |= REDIR_VARID_MASK;
-	    ncodes = 4;
+	    ncodes = 6;
 	}
 	else
-	    ncodes = 3;
+	    ncodes = 5;
 
 	/* If we ever to change the number of codes, we have to change
 	 * the definition of WC_REDIR_WORDS. */
@@ -1818,22 +1832,28 @@ par_redir(int *rp, char *idstring)
 	ecbuf[r] = WCB_REDIR(type);
 	ecbuf[r + 1] = fd1;
 
+	/*
+	 * r + 2: the HERE string we recover
+	 * r + 3: the HERE document terminator, raw
+	 * r + 4: the HERE document terminator, munged
+	 */
 	if (idstring)
-	    ecbuf[r + 3] = ecstrcode(idstring);
+	    ecbuf[r + 5] = ecstrcode(idstring);
 
-	for (hd = &hdocs; *hd; hd = &(*hd)->next);
+	for (hd = &hdocs; *hd; hd = &(*hd)->next)
+	    ;
 	*hd = zalloc(sizeof(struct heredocs));
 	(*hd)->next = NULL;
 	(*hd)->type = htype;
 	(*hd)->pc = r;
 	(*hd)->str = tokstr;
 
-	yylex();
+	zshlex();
 	return ncodes;
     }
     case REDIR_WRITE:
     case REDIR_WRITENOW:
-	if (tokstr[0] == Outang && tokstr[1] == Inpar)
+	if (tokstr[0] == OutangProc && tokstr[1] == Inpar)
 	    /* > >(...) */
 	    type = REDIR_OUTPIPE;
 	else if (tokstr[0] == Inang && tokstr[1] == Inpar)
@@ -1843,15 +1863,16 @@ par_redir(int *rp, char *idstring)
 	if (tokstr[0] == Inang && tokstr[1] == Inpar)
 	    /* < <(...) */
 	    type = REDIR_INPIPE;
-	else if (tokstr[0] == Outang && tokstr[1] == Inpar)
+	else if (tokstr[0] == OutangProc && tokstr[1] == Inpar)
 	    YYERROR(ecused);
 	break;
     case REDIR_READWRITE:
-	if ((tokstr[0] == Inang || tokstr[0] == Outang) && tokstr[1] == Inpar)
+	if ((tokstr[0] == Inang || tokstr[0] == OutangProc) &&
+	    tokstr[1] == Inpar)
 	    type = tokstr[0] == Inang ? REDIR_INPIPE : REDIR_OUTPIPE;
 	break;
     }
-    yylex();
+    zshlex();
 
     /* If we ever to change the number of codes, we have to change
      * the definition of WC_REDIR_WORDS. */
@@ -1876,10 +1897,12 @@ par_redir(int *rp, char *idstring)
 
 /**/
 void
-setheredoc(int pc, int type, char *str)
+setheredoc(int pc, int type, char *str, char *termstr, char *munged_termstr)
 {
     ecbuf[pc] = WCB_REDIR(type | REDIR_FROM_HEREDOC_MASK);
     ecbuf[pc + 2] = ecstrcode(str);
+    ecbuf[pc + 3] = ecstrcode(termstr);
+    ecbuf[pc + 4] = ecstrcode(munged_termstr);
 }
 
 /*
@@ -1894,7 +1917,7 @@ par_wordlist(void)
     while (tok == STRING) {
 	ecstr(tokstr);
 	num++;
-	yylex();
+	zshlex();
     }
     return num;
 }
@@ -1914,18 +1937,18 @@ par_nl_wordlist(void)
 	    ecstr(tokstr);
 	    num++;
 	}
-	yylex();
+	zshlex();
     }
     return num;
 }
 
 /*
- * condlex is yylex for normal parsing, but is altered to allow
+ * condlex is zshlex for normal parsing, but is altered to allow
  * the test builtin to use par_cond.
  */
 
 /**/
-void (*condlex) _((void)) = yylex;
+void (*condlex) _((void)) = zshlex;
 
 /*
  * cond	: cond_1 { SEPER } [ DBAR { SEPER } cond ]
@@ -2149,10 +2172,11 @@ par_cond_triple(char *a, char *b, char *c)
 	ecadd(ecnpats++);
     } else if ((b[0] == Equals || b[0] == '=') &&
                (b[1] == '~' || b[1] == Tilde) && !b[2]) {
+        /* We become an implicit COND_MODI but do not provide the first
+	 * item, it's skipped */
 	ecadd(WCB_COND(COND_REGEX, 0));
 	ecstr(a);
 	ecstr(c);
-	ecadd(ecnpats++);
     } else if (b[0] == '-') {
 	if ((t0 = get_cond_num(b + 1)) > -1) {
 	    ecadd(WCB_COND(t0 + COND_NT, 0));
@@ -2199,7 +2223,7 @@ yyerror(int noerr)
     int t0;
     char *t;
 
-    if ((t = dupstring(yytext)))
+    if ((t = dupstring(zshlextext)))
 	untokenize(t);
 
     for (t0 = 0; t0 != 20; t0++)
@@ -2428,10 +2452,15 @@ ecgetredirs(Estate s)
 	r->type = WC_REDIR_TYPE(code);
 	r->fd1 = *s->pc++;
 	r->name = ecgetstr(s, EC_DUP, NULL);
-	if (WC_REDIR_FROM_HEREDOC(code))
+	if (WC_REDIR_FROM_HEREDOC(code)) {
 	    r->flags = REDIRF_FROM_HEREDOC;
-	else
+	    r->here_terminator = ecgetstr(s, EC_DUP, NULL);
+	    r->munged_here_terminator = ecgetstr(s, EC_DUP, NULL);
+	} else {
 	    r->flags = 0;
+	    r->here_terminator = NULL;
+	    r->munged_here_terminator = NULL;
+	}
 	if (WC_REDIR_VARID(code))
 	    r->varid = ecgetstr(s, EC_DUP, NULL);
 	else
@@ -2736,7 +2765,7 @@ write_dump(int dfd, LinkList progs, int map, int hlen, int tlen)
 	fdsetflags(pre, ((map ? FDF_MAP : 0) | other));
 	fdsetother(pre, tlen);
 	strcpy(fdversion(pre), ZSH_VERSION);
-	write(dfd, pre, FD_PRELEN * sizeof(wordcode));
+	write_loop(dfd, (char *)pre, FD_PRELEN * sizeof(wordcode));
 
 	for (node = firstnode(progs); node; incnode(node)) {
 	    wcf = (WCFunc) getdata(node);
@@ -2757,11 +2786,11 @@ write_dump(int dfd, LinkList progs, int map, int hlen, int tlen)
 	    head.flags = fdhbldflags(wcf->flags, (tail - n));
 	    if (other)
 		fdswap((Wordcode) &head, sizeof(head) / sizeof(wordcode));
-	    write(dfd, &head, sizeof(head));
+	    write_loop(dfd, (char *)&head, sizeof(head));
 	    tmp = strlen(n) + 1;
-	    write(dfd, n, tmp);
+	    write_loop(dfd, n, tmp);
 	    if ((tmp &= (sizeof(wordcode) - 1)))
-		write(dfd, &head, sizeof(wordcode) - tmp);
+		write_loop(dfd, (char *)&head, sizeof(wordcode) - tmp);
 	}
 	for (node = firstnode(progs); node; incnode(node)) {
 	    prog = ((WCFunc) getdata(node))->prog;
@@ -2769,7 +2798,7 @@ write_dump(int dfd, LinkList progs, int map, int hlen, int tlen)
 		   sizeof(wordcode) - 1) / sizeof(wordcode);
 	    if (other)
 		fdswap(prog->prog, (((Wordcode) prog->strs) - prog->prog));
-	    write(dfd, prog->prog, tmp * sizeof(wordcode));
+	    write_loop(dfd, (char *)prog->prog, tmp * sizeof(wordcode));
 	}
 	if (other)
 	    break;
@@ -3084,6 +3113,8 @@ load_dump_file(char *dump, struct stat *sbuf, int other, int len)
 	return;
 
     fd = movefd(fd);
+    if (fd == -1)
+	return;
 
     if ((addr = (Wordcode) mmap(NULL, mlen, PROT_READ, MAP_SHARED, fd, off)) ==
 	((Wordcode) -1)) {
@@ -3414,6 +3445,7 @@ dump_autoload(char *nam, char *file, int on, Options ops, int func)
 	shf = (Shfunc) zshcalloc(sizeof *shf);
 	shf->node.flags = on;
 	shf->funcdef = mkautofn(shf);
+	shf->emulation = 0;
 	shfunctab->addnode(shfunctab, ztrdup(fdname(n) + fdhtail(n)), shf);
 	if (OPT_ISSET(ops,'X') && eval_autoload(shf, shf->node.nam, ops, func))
 	    ret = 1;
